"use strict";
/*
 * Deepkit Framework
 * Copyright (C) 2020 Deepkit UG
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.ConnectionMiddleware = exports.ConnectionWriter = exports.SimpleConnectionWriter = exports.isObservable = exports.AlreadyEncoded = void 0;
const collection_1 = require("./collection");
const core_1 = require("./core");
const core_rxjs_1 = require("@deepkit/core-rxjs");
const core_2 = require("@deepkit/core");
const type_1 = require("@deepkit/type");
const operators_1 = require("rxjs/operators");
class AlreadyEncoded {
    constructor(type, value) {
        this.type = type;
        this.value = value;
    }
}
exports.AlreadyEncoded = AlreadyEncoded;
function encodeValue(v, p, prefixMessage) {
    if (v instanceof AlreadyEncoded) {
        p = new type_1.PropertySchema('result');
        p.type = v.type;
        return {
            encoding: p.toJSON(),
            value: v.value,
        };
    }
    if (!p) {
        p = new type_1.PropertySchema('result');
        if (v !== undefined && v !== null) {
            p.setFromJSValue(v);
        }
    }
    try {
        return {
            encoding: p.toJSON(),
            value: type_1.jsonSerializer.serializeProperty(p, v),
        };
    }
    catch (error) {
        console.log('could not parse value', v, p);
        throw error;
    }
}
function isObservable(obj) {
    return obj && core_2.isFunction(obj.pipe) && core_2.isFunction(obj.subscribe) && core_2.isFunction(obj.toPromise);
}
exports.isObservable = isObservable;
class SimpleConnectionWriter {
    write(message) {
        throw new Error('Not implemented');
    }
    cancelMessage(messageId) {
    }
    complete(id) {
        if (!id)
            throw new Error('id required');
        this.write({ type: 'complete', id: id });
    }
    ack(id) {
        if (!id)
            throw new Error('id required');
        this.write({ type: 'ack', id: id });
    }
    sendError(id, errorObject, code) {
        const [entityName, error, stack] = core_1.getSerializedErrorPair(errorObject);
        this.write({ type: 'error', id: id, entityName, error, stack, code: error.code || code });
    }
}
exports.SimpleConnectionWriter = SimpleConnectionWriter;
class ConnectionWriter extends SimpleConnectionWriter {
    constructor(socket) {
        super();
        this.socket = socket;
        this.chunkIds = 0;
        this.cancelSending = {};
    }
    /**
     * This function handles backpressure and waits if necessary
     */
    async delayedWrite(data) {
        while (this.socket.bufferedAmount() > 1024 * 300) {
            await core_2.sleep(0.01);
        }
        let sent = await this.socket.send(data);
        let tries = 1;
        while (!sent) {
            console.log('sending failed, wait', data.substr(0, 20));
            await core_2.sleep(0.1);
            sent = await this.socket.send(data);
            tries++;
            if (tries > 10) {
                console.log('sending errored after', tries, 'tries', data.substr(0, 20));
                return;
            }
        }
        if (tries > 1) {
            console.log('sending succeeded after', tries, 'tries', data.substr(0, 20));
        }
    }
    /**
     * Queued writes will be canceled.
     * @param messageId
     */
    cancelMessage(messageId) {
        if (this.cancelSending[messageId]) {
            this.cancelSending[messageId]();
        }
    }
    // @stack()
    async write(message) {
        const json = JSON.stringify(message);
        const chunkSize = 1024 * 100;
        if (json.length > chunkSize) {
            const chunkId = this.chunkIds++;
            let sending = true;
            let id = 0;
            if (message.type === 'channel' || message.type === 'entity/patch' || message.type === 'entity/remove'
                || message.type === 'entity/removeMany' || message.type === 'entity/update' || message.type === 'push-message') {
            }
            else {
                id = message.id;
            }
            if (id) {
                this.cancelSending[id] = () => {
                    sending = false;
                };
            }
            let position = 0;
            await this.delayedWrite('@batch-start:' + (message['id'] || 0) + ':' + chunkId + ':' + json.length);
            while (sending && position * chunkSize < json.length) {
                const chunk = json.substr(position * (chunkSize), chunkSize);
                position++;
                //maybe we should add here additional delay after x MB, so other connection/messages get their time as well.
                // console.log('sent chunk', chunkId, position, chunk.substr(0, 20));
                await this.delayedWrite('@batch:' + chunkId + ':' + chunk);
            }
            if (sending) {
                // console.log('chunk done', chunkId, position);
                await this.delayedWrite('@batch-end:' + chunkId);
            }
            if (id) {
                delete this.cancelSending[id];
            }
        }
        else {
            await this.delayedWrite(json);
        }
    }
    complete(id) {
        if (!id)
            throw new Error('id required');
        this.write({ type: 'complete', id: id });
    }
    ack(id) {
        if (!id)
            throw new Error('id required');
        this.write({ type: 'ack', id: id });
    }
    sendError(id, errorObject, code) {
        const [entityName, error, stack] = core_1.getSerializedErrorPair(errorObject);
        this.write({ type: 'error', id: id, entityName, error, stack, code: error.code || code });
    }
}
exports.ConnectionWriter = ConnectionWriter;
class ConnectionMiddleware {
    constructor() {
        this.collectionSubscriptions = {};
        this.collections = {};
        this.subjectSubscriptions = {};
        this.observables = {};
        this.entitySent = {};
    }
    destroy() {
        for (const sub of core_2.each(this.collectionSubscriptions)) {
            sub.unsubscribe();
        }
        for (const sub of core_2.each(this.subjectSubscriptions)) {
            sub.unsubscribe();
        }
        for (const ob of core_2.each(this.observables)) {
            for (const sub of core_2.each(ob.subscriber)) {
                sub.unsubscribe();
            }
        }
    }
    async messageIn(message, writer) {
        // console.log('messageIn', message);
        if (message.name === 'entity/unsubscribe') {
            const sent = this.entitySent[message.forId];
            if (!sent) {
                throw new Error(`Entity not sent for message ${message.id}`);
            }
            writer.ack(message.id);
            return;
        }
        if (message.name === 'subject/unsubscribe') {
            const sent = this.subjectSubscriptions[message.forId];
            if (!sent) {
                throw new Error(`Subject not subscribed ${message.forId}`);
            }
            await sent.unsubscribe();
            writer.ack(message.id);
            return;
        }
        if (message.name === 'collection/unsubscribe') {
            if (this.collectionSubscriptions[message.forId]) {
                this.collectionSubscriptions[message.forId].unsubscribe();
            }
            writer.ack(message.id);
            return;
        }
        if (message.name === 'collection/pagination') {
            if (this.collections[message.forId]) {
                //happens when the client sent pagination changes.
                // console.log('client send pagination updates', message);
                this.collections[message.forId].pagination.setSort(message.sort);
                this.collections[message.forId].pagination.setPage(message.page);
                this.collections[message.forId].pagination.setItemsPerPage(message.itemsPerPage);
                this.collections[message.forId].pagination.setParameters(message.parameters);
                this.collections[message.forId].pagination.event.next({ type: 'client:apply' });
            }
            writer.ack(message.id);
            return;
        }
        if (message.name === 'observable/subscribe') {
            if (!this.observables[message.forId]) {
                throw new Error('No observable registered.');
            }
            if (this.observables[message.forId].subscriber[message.subscribeId]) {
                throw new Error('Subscriber already registered.');
            }
            this.observables[message.forId].subscriber[message.subscribeId] = this.observables[message.forId].observable.subscribe((next) => {
                const { encoding, value } = encodeValue(next, this.observables[message.forId].p, `${this.observables[message.forId].prefix} observable next`);
                writer.write({
                    type: 'next/observable',
                    id: message.forId,
                    subscribeId: message.subscribeId,
                    encoding: encoding,
                    next: value
                });
            }, (errorObject) => {
                const [entityName, error, stack] = core_1.getSerializedErrorPair(errorObject);
                writer.write({
                    type: 'error/observable',
                    id: message.forId,
                    entityName, error, stack,
                    subscribeId: message.subscribeId
                });
            }, () => {
                writer.write({
                    type: 'complete/observable',
                    id: message.forId,
                    subscribeId: message.subscribeId
                });
            });
            writer.ack(message.forId);
        }
        if (message.name === 'observable/unsubscribe') {
            if (!this.observables[message.forId]) {
                throw new Error('No observable registered.');
            }
            if (!this.observables[message.forId].subscriber[message.subscribeId]) {
                throw new Error('Subscriber already unsubscribed.');
            }
            this.observables[message.forId].subscriber[message.subscribeId].unsubscribe();
            delete this.observables[message.forId].subscriber[message.subscribeId];
            if (core_2.getObjectKeysSize(this.observables[message.forId].subscriber) === 0) {
                delete this.observables[message.forId];
            }
            writer.ack(message.forId);
        }
    }
    async actionMessageOut(message, result, propertySchema, //of the method return type
    controllerName, actionName, writer) {
        // console.log('messageOut', {
        //     EntitySubject: result instanceof EntitySubject,
        //     StreamBehaviorSubject: result instanceof StreamBehaviorSubject,
        //     Collection: result instanceof Collection,
        //     Observable: result instanceof Observable,
        //     className: getClassName(result),
        // }, result);
        if (result instanceof Promise) {
            throw new Error('Promise not supported as actionMessageOut');
        }
        const prefix = `${controllerName}::${actionName}`;
        if (result instanceof core_1.EntitySubject) {
            const item = result.getValue();
            if (undefined === item) {
                writer.write({
                    type: 'type',
                    id: message.id,
                    returnType: 'entity',
                    entityName: undefined,
                    item: undefined,
                });
                return;
            }
            const entityName = type_1.getEntityName(item.constructor);
            this.entitySent[message.id] = {
                classType: item.constructor,
                id: item.id,
            };
            writer.write({
                type: 'type',
                id: message.id,
                returnType: 'entity',
                entityName: entityName,
                item: entityName ? type_1.jsonSerializer.for(item.constructor).serialize(item) : item
            });
            writer.complete(message.id);
            //no further subscribes/messages necessary since the 'entity' channel handles updating.
            //this means, once this entity is registered in entity-storage, we automatically push changes of this entity.
        }
        else if (result instanceof core_1.StreamBehaviorSubject) {
            const item = result.getValue();
            const { encoding, value } = encodeValue(item, propertySchema.getTemplateArg(0), `${prefix} subject initial`);
            writer.write({
                type: 'type',
                id: message.id,
                returnType: 'subject',
                encoding: encoding,
                data: value,
            });
            this.subjectSubscriptions[message.id] = new core_rxjs_1.Subscriptions(async () => {
                writer.cancelMessage(message.id);
                await result.unsubscribe();
                delete this.subjectSubscriptions[message.id];
            });
            this.subjectSubscriptions[message.id].add = result.appendSubject.subscribe((append) => {
                const { encoding, value } = encodeValue(append, propertySchema.getTemplateArg(0), `${prefix} subject append`);
                writer.write({
                    type: 'append/subject',
                    id: message.id,
                    encoding: encoding,
                    append: value
                });
            });
            //we sent already the first initial value, since its a BehaviorSubject we skip the first
            this.subjectSubscriptions[message.id].add = result.pipe(operators_1.skip(1)).subscribe((next) => {
                const { encoding, value } = encodeValue(next, propertySchema.getTemplateArg(0), `${prefix} subject next`);
                writer.write({
                    type: 'next/subject',
                    id: message.id,
                    encoding: encoding,
                    next: value,
                });
            }, async (error) => {
                writer.sendError(message.id, error);
                await this.subjectSubscriptions[message.id].unsubscribe();
            }, async () => {
                writer.complete(message.id);
                await this.subjectSubscriptions[message.id].unsubscribe();
            });
        }
        else if (result instanceof collection_1.Collection) {
            const collection = result;
            writer.write({
                type: 'type',
                id: message.id,
                returnType: 'collection',
                pagination: {
                    active: collection.pagination.isActive(),
                    itemsPerPage: collection.pagination.getItemsPerPage(),
                    page: collection.pagination.getPage(),
                    total: collection.pagination.getTotal(),
                    sort: collection.pagination.getSort(),
                    parameters: collection.pagination.getParameters(),
                },
                entityName: type_1.getEntityName(collection.classType)
            });
            let nextValue;
            const serializer = type_1.jsonSerializer.for(collection.classType);
            const items = collection instanceof collection_1.JSONObjectCollection
                ? collection.all()
                : collection.all().map(v => serializer.serialize(v));
            nextValue = {
                type: 'set',
                total: collection.count(),
                items: items
            };
            writer.write({ type: 'next/collection', id: message.id, next: nextValue });
            if (this.collectionSubscriptions[message.id]) {
                throw new Error('Collection already subscribed');
            }
            this.collectionSubscriptions[message.id] = new core_rxjs_1.Subscriptions(() => {
                writer.cancelMessage(message.id);
                collection.unsubscribe();
                delete this.collections[message.id];
                delete this.collectionSubscriptions[message.id];
            });
            this.collections[message.id] = collection;
            this.collectionSubscriptions[message.id].add = collection.subscribe(() => {
            }, (error) => {
                writer.sendError(message.id, error);
            }, () => {
                writer.complete(message.id);
            });
            const sendPagination = () => {
                nextValue = {
                    type: 'pagination',
                    event: {
                        type: 'server:change',
                        order: collection.pagination.getSort(),
                        itemsPerPage: collection.pagination.getItemsPerPage(),
                        page: collection.pagination.getPage(),
                        total: collection.pagination.getTotal(),
                        parameters: collection.pagination.getParameters(),
                    }
                };
                writer.write({ type: 'next/collection', id: message.id, next: nextValue });
            };
            this.collectionSubscriptions[message.id].add = collection.pagination.event.subscribe((event) => {
                if (event.type.startsWith('server:')) {
                    writer.write({ type: 'next/collection', id: message.id, next: { type: 'pagination', event: event } });
                }
                //happens when a query change or external (client) pagination change resulted in some pagination parameter changes (like total)
                //so we send again the current state to the client.
                if (event.type === 'internal_server_change') {
                    sendPagination();
                }
                //happens when the controller which created the collection changed the pagination. we then send the current state to the client.
                if (event.type === 'apply') {
                    sendPagination();
                }
            });
            this.collectionSubscriptions[message.id].add = collection.event.subscribe((event) => {
                if (event.type === 'add') {
                    const item = collection instanceof collection_1.JSONObjectCollection
                        ? event.item
                        : type_1.jsonSerializer.for(collection.classType).serialize(event.item);
                    nextValue = { type: 'add', item: item };
                    writer.write({ type: 'next/collection', id: message.id, next: nextValue });
                }
                if (event.type === 'removeMany') {
                    nextValue = { type: 'removeMany', ids: event.ids };
                    writer.write({ type: 'next/collection', id: message.id, next: nextValue });
                }
                if (event.type === 'remove') {
                    nextValue = { type: 'remove', id: event.id };
                    writer.write({ type: 'next/collection', id: message.id, next: nextValue });
                }
                if (event.type === 'batch/start' || event.type === 'batch/end') {
                    writer.write({ type: 'next/collection', id: message.id, next: event });
                }
                if (event.type === 'sort') {
                    nextValue = { type: 'sort', ids: event.ids };
                    writer.write({ type: 'next/collection', id: message.id, next: nextValue });
                }
                if (event.type === 'set') {
                    //consider batching the items, so we don't block the connection stack
                    //when we have thousand of items and we get a nice loading bar.
                    const serializer = type_1.jsonSerializer.for(collection.classType);
                    const items = collection instanceof collection_1.JSONObjectCollection
                        ? event.items
                        : event.items.map(v => serializer.serialize(v));
                    nextValue = {
                        type: 'set',
                        total: event.items.length,
                        items: items
                    };
                    writer.write({ type: 'next/collection', id: message.id, next: nextValue });
                }
            });
        }
        else if (isObservable(result)) {
            writer.write({ type: 'type', id: message.id, returnType: 'observable' });
            this.observables[message.id] = { observable: result, subscriber: {}, p: propertySchema.getTemplateArg(0), prefix: prefix };
        }
        else {
            const v = encodeValue(result, propertySchema, `${prefix} result`);
            writer.write({ type: 'next/json', id: message.id, encoding: v.encoding, next: v.value });
        }
    }
}
exports.ConnectionMiddleware = ConnectionMiddleware;
//# sourceMappingURL=connection.js.map